/**
 * @file wqueue.c
 * Work queue for multi-threaded applications
 * @author Marko Mäkelä (marko.makela@iki.fi)
 */

/* Copyright © 2007 Marko Mäkelä.

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

#include "wqueue.h"

/** Worker thread.
 * @param wq	work queue
 * @param proc	work procedure
 * @param ctx	context of work procedure
 */
void*
wqueue_worker (struct wqueue* wq,
	       int (*proc) (struct dlist_item*, void*),
	       void* ctx)
{
  pthread_mutex_lock (&wq->mutex);
  wq->n_working++;

  for (;;) {
    struct dlist_item* work;
    int status;

    /* Wait for a work request */
    while (!wq->work.size) {
      pthread_cond_signal (&wq->one_finished);

      if (!--wq->n_working)
	pthread_cond_signal (&wq->all_finished);

      if (wq->stop)
	break;

      wq->n_waiting++;
      pthread_cond_wait (&wq->work_wait, &wq->mutex);

      wq->n_waiting--;
      wq->n_working++;
    }

    if (wq->stop)
      break;

    dlist_remove_first (&wq->work, work);
    pthread_mutex_unlock (&wq->mutex);

    status = (*proc) (work, ctx);

    pthread_mutex_lock (&wq->mutex);

    if (status)
      wq->stop = 1;
  }

  pthread_mutex_unlock (&wq->mutex);
  return 0;
}


/** Enqueue work.
 * @param wq	work queue
 * @param work	work to enqueue
 * @param limit	upper limit of the length of the work queue
 * @return	zero on success, nonzero on failure
 */
int
wqueue_enqueue (struct wqueue* wq,
		struct dlist_item* work,
		int limit)
{
  /* Prepare to enqueue work. */
  pthread_mutex_lock (&wq->mutex);

  /* Wait for the workers to process some requests. */
  while (wq->work.size > limit) {
    if (wq->stop) {
      pthread_mutex_unlock (&wq->mutex);
      return 1;
    }

    /* Signal any pending workers to process a request. */
    if (wq->n_waiting)
      pthread_cond_signal(&wq->work_wait);

    /* Wait for one worker to finish. */
    pthread_cond_wait (&wq->one_finished, &wq->mutex);
  }

  /* Enqueue work. */
  dlist_add_last (&wq->work, work);

  if (wq->n_waiting)
    pthread_cond_signal(&wq->work_wait);
  pthread_mutex_unlock (&wq->mutex);
  return 0;
}

/** Wait for all enqueued work to complete.
 * @param wq	work queue
 * @return	zero on success; nonzero on fatal error
 */
int
wqueue_wait_empty (struct wqueue* wq)
{
  int stop;

  pthread_mutex_lock (&wq->mutex);

  /* Wait for all workers to finish. */
  while (wq->n_working || (!wq->stop && wq->work.size))
    pthread_cond_wait (&wq->all_finished, &wq->mutex);

  stop = wq->stop;

  pthread_mutex_unlock (&wq->mutex);

  return stop;
}

/** Wait for all workers to terminate.
 * @param wq	work queue
 * @param stop	nonzero=signal all workers to terminate
 * @param clean	procedure to clean up any unprocessed work items
 */
void
wqueue_terminate (struct wqueue* wq,
		  int stop,
		  void (*clean) (struct dlist_item*))
{
  pthread_mutex_lock (&wq->mutex);

  if (stop) {
    /* Ask the workers to terminate immediately. */
    wq->stop = stop;
    pthread_cond_broadcast (&wq->work_wait);
  }

  /* Wait for all workers to finish. */
  while (wq->n_working || (!wq->stop && wq->work.size))
    pthread_cond_wait (&wq->all_finished, &wq->mutex);

  /* Tell all workers to stop waiting, and wait for them to finish. */
  wq->stop = 1;
  pthread_cond_broadcast (&wq->work_wait);

  while (wq->n_waiting)
    pthread_cond_wait (&wq->all_finished, &wq->mutex);

  /* Discard any unprocessed work. */
  for (;;) {
    struct dlist_item* work;
    dlist_remove_first (&wq->work, work);
    if (!work)
      break;
    (*clean) (work);
  }

  pthread_mutex_unlock (&wq->mutex);
}
