/**
 * @file wqueue.h
 * Work queue for multi-threaded applications
 * @author Marko Mäkelä (marko.makela@iki.fi)
 */

/* Copyright © 2007 Marko Mäkelä.

   This file is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

#include <pthread.h>
#include "dlist.h"

/** Work queue structure. */
struct wqueue
{
  /** Mutual exclusion device for protecting this structure */
  pthread_mutex_t mutex;
  /** Workers waiting for work */
  pthread_cond_t work_wait;
  /** Boss waiting for all workers to finish */
  pthread_cond_t all_finished;
  /** Boss waiting for one worker to finish */
  pthread_cond_t one_finished;
  /** Number of workers working */
  int n_working;
  /** Number of workers waiting for work */
  int n_waiting;
  /** Pending request to stop all work */
  int stop;
  /** The work queue. */
  struct dlist_base work;
};

/** Work queue initiliazer */
#define WQUEUE_INITITIALIZER {			\
  PTHREAD_MUTEX_INITIALIZER,			\
  PTHREAD_COND_INITIALIZER,			\
  PTHREAD_COND_INITIALIZER,			\
  PTHREAD_COND_INITIALIZER,			\
  0, 0, 0,					\
  DLIST_BASE_INITIALIZER			\
}

/** Worker thread.
 * @param wq	work queue
 * @param proc	work procedure
 * @param ctx	context of work procedure
 */
void*
wqueue_worker (struct wqueue* wq,
	       int (*proc) (struct dlist_item*, void*),
	       void* ctx);

/** Enqueue work.
 * @param wq	work queue
 * @param work	work to enqueue
 * @param limit	upper limit of the length of the work queue
 * @return	zero on success, nonzero on failure
 */
int
wqueue_enqueue (struct wqueue* wq,
		struct dlist_item* work,
		int limit);

/** Wait for all enqueued work to complete.
 * @param wq	work queue
 * @return	zero on success; nonzero on fatal error
 */
int
wqueue_wait_empty (struct wqueue* wq);

/** Wait for all workers to terminate.
 * @param wq	work queue
 * @param stop	nonzero=signal all workers to terminate
 * @param clean	procedure to clean up any unprocessed work items
 */
void
wqueue_terminate (struct wqueue* wq,
		  int stop,
		  void (*clean) (struct dlist_item*));
